/****************************************************************************************/
/*											*/
/* This program is free software; you can redistribute it and/or modify it under the	*/
/* terms of the GNU General Public License as published by the Free Software		*/
/* Foundation; either version 2 of the License, or (at your option) any later version.	*/
/*											*/
/* This program is distributed in the hope that it will be useful, but WITHOUT ANY	*/
/* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A	*/
/* PARTICULAR PURPOSE. See the GNU General Public License for more details.		*/
/*											*/
/* You should have received a copy of the GNU General Public License along with this	*/
/* program; (See "COPYING"). If not, If not, see <http://www.gnu.org/licenses/>.        */
/*											*/
/*--------------------------------------------------------------------------------------*/
/*											*/
/*  Copyright   Joerg Anders, TU Chemnitz, Fakultaet fuer Informatik, GERMANY           */
/*		ja@informatik.tu-chemnitz.de						*/
/*											*/
/*											*/
/****************************************************************************************/

#include "pangocairotext.h"
#include "resource.h"

#define TEXT_CURSOR_WIDTH 0.04
#define TEXT_CURSOR_HEIGHT 0.3

#define MAX_TEXT_LENGTH 10

#define PANGO_INTERNAL_FONT_FACTOR 1024.0

#define INTERNAL_Y_OFFSET -0.2
#define SPEC_LETTERS_SCALE 600.0

#define CODA_X_OFFSET 0.1
#define CODA_Y_OFFSET 0.21

#define SEGNO_X_OFFSET 0.1
#define SEGNO_Y_OFFSET 0.2

#define SIGN_X_OFFSET 0.1
#define SIGN_Y_OFFSET 0.2

#define FLAT_X_OFFSET 0.0
#define FLAT_Y_OFFSET 0.25

NedPangoCairoText::NedPangoCairoText(GdkDrawable *d, char *text, const char* font_family, PangoStyle style, PangoWeight weight, double size, double zoom, double scale) :
m_display_layout(NULL), m_measure_layout(NULL), m_drawable(d), m_hidden_hyphen(false), m_text(NULL), m_measure_text(NULL), m_font_family(font_family), m_style(style), m_weight(weight), m_size(size * PANGO_INTERNAL_FONT_FACTOR), m_width(0), m_height(0), 
	m_cursor_idx(0), m_spec_letters(NULL) {
	if (text == NULL || strlen(text) < 1) {
		NedResource::Abort("NedPangoCairoText::NedPangoCairoText");
	}
	m_text = inspect_text(text);
	m_measure_text = strdup(text);
	if (m_text[strlen(m_text) - 1] == '-') {
		m_hidden_hyphen = true;
		m_text[strlen(m_text) - 1] = '\0';
	}
	m_drawable = d;
	m_font_family = font_family; 
	m_style = style;
	m_weight = weight;
	createLayout(zoom, scale);
	m_cursor_idx = strlen(m_text);
	computeCursorPosition();
}
	
NedPangoCairoText::~NedPangoCairoText() {
	GList *lptr;

	if (m_display_layout != NULL) {
		g_object_unref (m_display_layout);
		m_display_layout = NULL;
	}
	if (m_measure_layout != NULL) {
		g_object_unref (m_measure_layout);
		m_measure_layout = NULL;
	}
	if (m_text != NULL) {
		free(m_text);
		m_text = NULL;
	}
	for (lptr = g_list_first(m_spec_letters); lptr; lptr = g_list_next(lptr)) {
		delete (specLetter *) lptr->data;
	}
	g_list_free(m_spec_letters);
}

char *NedPangoCairoText::getText() {
	return m_measure_text;
}

double NedPangoCairoText::getInternalFactor() {
	return PANGO_INTERNAL_FONT_FACTOR;
}

void NedPangoCairoText::adjust(GdkDrawable *d, double zoom, double scale) {
	m_drawable = d;
	createLayout(zoom, scale);
	computeDimensions();
	computeCursorPosition();
}

NedPangoCairoText *NedPangoCairoText::clone() {
	return new NedPangoCairoText(m_drawable, m_measure_text, m_font_family, m_style, m_weight, m_size / PANGO_INTERNAL_FONT_FACTOR, NedResource::getStartZoomScale(), 1.0);
}

	

void NedPangoCairoText::setText(char *text, double zoom) {
	GList *lptr;
	if (m_text != NULL) {
		free(m_text);
		m_text = NULL;
		free(m_measure_text);
		m_measure_text = NULL;
	}
	for (lptr = g_list_first(m_spec_letters); lptr; lptr = g_list_next(lptr)) {
		delete (specLetter *) lptr->data;
	}
	g_list_free(m_spec_letters);
	m_spec_letters = NULL;
	m_hidden_hyphen = false;
	if (text != NULL && strlen(text) > 0) {
		m_text = inspect_text(text);
		m_measure_text = strdup(text);
		if (m_text[strlen(m_text) - 1] == '-') {
			m_hidden_hyphen = true;
			m_text[strlen(m_text) - 1] = '\0';
		}
	}
	if (m_text == NULL) {
		pango_layout_set_text (m_display_layout, " ", -1);
	}
	else if (strlen(m_text) < 1) {
		pango_layout_set_text (m_display_layout, " ", -1);
	}
	else {
		pango_layout_set_text (m_display_layout, m_text, -1);
		pango_layout_set_text (m_measure_layout, m_measure_text, -1);
	}
	computeDimensions();
	computeCursorPosition();
}

void NedPangoCairoText::setZoom(double zoom, double scale) {
	PangoFontDescription *desc;
	cairo_t *cr = gdk_cairo_create (m_drawable);
	desc = pango_font_description_new();
	pango_font_description_set_family(desc, m_font_family);
	pango_font_description_set_style(desc, m_style);
	pango_font_description_set_weight(desc, m_weight);
	pango_font_description_set_absolute_size (desc, (int) (m_size * zoom * scale));
	pango_layout_set_font_description (m_display_layout, desc);
	pango_layout_set_font_description (m_measure_layout, desc);
	pango_cairo_update_layout (cr, m_display_layout);
	pango_cairo_update_layout (cr, m_measure_layout);
	pango_layout_get_size (m_measure_layout, &m_width, &m_height);
	computeCursorPosition();
	pango_font_description_free (desc);
	cairo_destroy (cr);
}

	
void NedPangoCairoText::createLayout(double zoom, double scale) {
	PangoFontDescription *desc;
	cairo_t *cr = gdk_cairo_create (m_drawable);

	m_display_layout = pango_cairo_create_layout (cr);
	m_measure_layout = pango_cairo_create_layout (cr);
	if (m_text == NULL) {
		pango_layout_set_text (m_display_layout, " ", -1);
		pango_layout_set_text (m_measure_layout, " ", -1);
	}
	else if (strlen(m_text) < 1) {
		pango_layout_set_text (m_display_layout, " ", -1);
		pango_layout_set_text (m_measure_layout, " ", -1);
	}
	else {
		pango_layout_set_text (m_display_layout, m_text, -1);
		pango_layout_set_text (m_measure_layout, m_measure_text, -1);
	}
	desc = pango_font_description_new();
	pango_font_description_set_family(desc, m_font_family);
	pango_font_description_set_style(desc, m_style);
	pango_font_description_set_weight(desc, m_weight);
	pango_font_description_set_absolute_size (desc, (int) (m_size * zoom * scale));
	pango_layout_set_font_description (m_display_layout, desc);
	pango_layout_set_font_description (m_measure_layout, desc);
	pango_cairo_update_layout (cr, m_display_layout);
	pango_cairo_update_layout (cr, m_measure_layout);
	pango_layout_get_size (m_measure_layout, &m_width, &m_height);
	pango_font_description_free (desc);
	cairo_destroy (cr);
}

bool NedPangoCairoText::textDiffers(char *other_text) {
	if (other_text == NULL || strlen(other_text) < 1) {
		return !(m_text == NULL || strlen(m_text) < 1);
	}
	if (m_measure_text == NULL || strlen(m_measure_text) < 1) return true;
	return strcmp(m_measure_text, other_text);
}

double NedPangoCairoText::getFontSize() {
	return m_size;
}

void NedPangoCairoText::draw(cairo_t *cr, double xpos, double ypos, double zoom, double scale) {
	GList *lptr;
	specLetter *let;
	cairo_matrix_t m;
	double scale_factor =  m_size / SPEC_LETTERS_SCALE * scale;
	double x_shift, y_shift;
	cairo_new_path(cr);
	cairo_move_to(cr, xpos, ypos + INTERNAL_Y_OFFSET * zoom);
	pango_cairo_show_layout (cr, m_display_layout);
	if (m_spec_letters != NULL) {
		cairo_get_matrix(cr, &m);
		cairo_scale(cr, scale_factor, scale_factor);
		for (lptr = g_list_first(m_spec_letters); lptr; lptr = g_list_next(lptr)) {
			let = (specLetter *) lptr->data;
			y_shift = (double) let->pos.y / PANGO_INTERNAL_FONT_FACTOR + let->yoffs;
			x_shift = xpos + (double) let->pos.x / PANGO_INTERNAL_FONT_FACTOR + let->xoffs * zoom;
			y_shift = ypos + (double) let->pos.y / PANGO_INTERNAL_FONT_FACTOR + (INTERNAL_Y_OFFSET + let->yoffs) * zoom;
			cairo_move_to(cr, x_shift / scale_factor, y_shift / scale_factor);
			cairo_show_text(cr, let->letter);
		}
		cairo_set_matrix(cr, &m);
	}
	cairo_stroke(cr);
}

void NedPangoCairoText::drawCursor(cairo_t *cr, double xpos, double ypos, double zoom_factor) {
	cairo_set_source_rgb (cr, 1.0, 0.0, 0.0);
	cairo_new_path(cr);
	cairo_rectangle (cr, xpos + (double) m_cursorpos.x / PANGO_INTERNAL_FONT_FACTOR,
		 ypos + m_cursorpos.y / PANGO_INTERNAL_FONT_FACTOR,
	TEXT_CURSOR_WIDTH * zoom_factor, TEXT_CURSOR_HEIGHT * zoom_factor);
	cairo_fill (cr);
	cairo_set_source_rgb (cr, 0.0, 0.0, 0.0);
}

void NedPangoCairoText::computeDimensions() {
	cairo_t *cr = gdk_cairo_create (m_drawable);
	pango_cairo_update_layout (cr, m_display_layout);
	pango_layout_get_size (m_measure_layout, &m_width, &m_height);
	cairo_destroy (cr);
}

void NedPangoCairoText::insertAtCursorPosition(char *text, double zoom) {
	char str1[1024], str2[1024];
	//text = "ע";
	if (text == NULL || strlen(text) < 1) return;
	bool RTL = pango_find_base_dir(m_text, -1) == PANGO_DIRECTION_RTL;
	if (m_cursor_idx == 0) {
		strcpy(str1, text);
		strcat(str1, m_text);
		setText(str1, zoom);
		m_cursor_idx = 0;
		if (RTL) moveCursor(-1);
		computeCursorPosition();
		return;
	}
	if (m_cursor_idx == strlen(m_measure_text)) {
		strcpy(str1, m_text);
		strcat(str1, text);
		setText(str1, zoom);
	 	m_cursor_idx = strlen(m_measure_text);
		m_cursorpos.x = RTL ? 0 : m_width;
		computeDimensions();
		return;
	}
	strncpy(str1, m_text, m_cursor_idx);
	str1[m_cursor_idx] = '\0';
	strcpy(str2, m_text + m_cursor_idx);
	strcat(str1, text);
	strcat(str1, str2);
	setText(str1, zoom);
	computeDimensions();
	moveCursor(RTL ? -1 : 1);
}

bool NedPangoCairoText::removeAtCursorPosition(double zoom) {
	char str1[1024], str2[1024];
	unsigned int old_cursor_idx = m_cursor_idx;
	if (m_cursor_idx == 0) return false;
	if (g_utf8_strlen(m_measure_text, -1) <= 1) return false;
	bool RTL = pango_find_base_dir(m_text, -1) == PANGO_DIRECTION_RTL;

	if (RTL) {
		moveCursor(1);
		if (old_cursor_idx < strlen(m_measure_text)) {
			strcpy(str2, m_text + old_cursor_idx);
		}
		else {
			str2[0] = '\0';
		}
		strncpy(str1, m_text, m_cursor_idx);
		str1[m_cursor_idx] = '\0';
		strcat(str1, str2);
	}
	else {
		moveCursor(-1);
		if (old_cursor_idx < strlen(m_measure_text)) {
			strcpy(str2, m_text + old_cursor_idx);
		}
		else {
			str2[0] = '\0';
		}
		strncpy(str1, m_text, m_cursor_idx);
		str1[m_cursor_idx] = '\0';
		strcat(str1, str2);
	}
	setText(str1, zoom);
	computeDimensions();
	return true;
}

void NedPangoCairoText::setCursorToLastPos() {
	if (m_text == NULL) return;
	bool RTL = (pango_find_base_dir(m_text, -1) == PANGO_DIRECTION_RTL);

	m_cursor_idx = strlen(m_measure_text);
	if (RTL) {
		m_cursor_idx = 0;
		computeCursorPosition();
	}
	else {
		m_cursorpos.x = m_width;
		m_cursor_idx = strlen(m_text);
	}
}

void NedPangoCairoText::setCursorToFirstPos() {
	if (m_text == NULL) return;
	bool RTL = (pango_find_base_dir(m_text, -1) == PANGO_DIRECTION_RTL);
	if (RTL) {
		m_cursor_idx = strlen(m_text);
		computeCursorPosition();
	}
	else {
		m_cursor_idx = 0;
		computeCursorPosition();
	}
}

bool NedPangoCairoText::moveCursor(int dist) {
	int trailing;
	gchar *cptr;
	int newcursor_idx;
	unsigned int i;
	unsigned int ll = strlen(m_measure_text);
	unsigned int llutf8;
	bool RTL = pango_find_base_dir(m_text, -1) == PANGO_DIRECTION_RTL;

	if (m_cursor_idx == ll) {
		if (dist > 0 && !RTL || dist < 0 && RTL) {
			return false;
		}
		if (dist == -1 && !RTL || dist == 1 && RTL) {
			cptr = m_text;
			llutf8 = g_utf8_strlen(m_measure_text, -1);
			for (i = 0; i < llutf8 - 1; i++) {
				cptr = g_utf8_next_char(cptr);
			}
			m_cursor_idx = cptr - m_text;
			computeCursorPosition();
			return true;
		}
		return false;
	}
	pango_layout_move_cursor_visually(m_measure_layout, TRUE, m_cursor_idx, 0, dist, &newcursor_idx, &trailing);
	if (newcursor_idx < 0) {
		return false;
	}
	if (trailing > 0) {
		m_cursor_idx = ll;
		if (RTL) {
			m_cursorpos.x = 0;
		}
		else {
			m_cursorpos.x = m_width;
		}
		return true;
	}
		
	m_cursor_idx = newcursor_idx;
	computeCursorPosition();
	return true;
}

double NedPangoCairoText::getWidth() {
	return (double) m_width / PANGO_INTERNAL_FONT_FACTOR;
}
double NedPangoCairoText::getHeight() {
	return (double) m_height / PANGO_INTERNAL_FONT_FACTOR;
}

void NedPangoCairoText::computeCursorPosition() {
	GList *lptr;
	int i;
	PangoRectangle pos2;
	pango_layout_get_cursor_pos(m_measure_layout, m_cursor_idx, &m_cursorpos, &pos2);
	for (i = 0, lptr = g_list_first(m_spec_letters); lptr; lptr = g_list_next(lptr), i++) {
		pango_layout_get_cursor_pos(m_display_layout, ((specLetter *) lptr->data)->idx,
			 &(((specLetter *) lptr->data)->pos), &pos2);
		((specLetter *) lptr->data)->pos.x = ((specLetter *) lptr->data)->pos.x;
	}
}

char *NedPangoCairoText::inspect_text(char *text) {
	char Str[1024];
	char *cptr, *cptr2, *cptr3;
	specLetter *let;

	strcpy(Str, text);
	cptr = Str;
	while (*cptr != '\0') {
		let = NULL;
		if (*cptr == '\\') {
			cptr2 = g_utf8_next_char(cptr);
			if (*cptr2 != '\0') {
				cptr3 = g_utf8_next_char(cptr2);
				switch (*cptr2) {
					case 'O': let = new specLetter("\160",
							 cptr - Str,
							 CODA_X_OFFSET, CODA_Y_OFFSET);
						  break;
					case 'S': let = new specLetter(
						  	"\161",
						  	cptr - Str,
						  	SEGNO_X_OFFSET, SEGNO_Y_OFFSET);
						  break;
					case '#': let = new specLetter(
						  	"\100",
						  	cptr - Str,
						  	SIGN_X_OFFSET, SIGN_Y_OFFSET);
						  break;
					case 'b': let = new specLetter(
						  	"\121",
						  	cptr - Str,
						  	FLAT_X_OFFSET,
						  	FLAT_Y_OFFSET);
						  break;
					case 'n': let = new specLetter(
						  	"\122",
						  	cptr - Str,
						  	SIGN_X_OFFSET,
						  	SIGN_Y_OFFSET);
						  break;
				}
				if (let != NULL) {
					*cptr = ' ';
					*cptr2 = ' ';
					cptr = cptr3;
					m_spec_letters = g_list_append(m_spec_letters, let);
				}
				else {
					cptr = g_utf8_next_char(cptr);
				}
			}
			else {
				cptr = g_utf8_next_char(cptr);
			}
		}
		else {
			cptr = g_utf8_next_char(cptr);
		}
	}
	return strdup(Str);
}
	
NedPangoCairoText::specLetter::specLetter(const char *let, int ix, double xo, double yo) :
		letter(let), idx(ix), xoffs(xo), yoffs(yo) {}

